﻿using UnityEngine;
using System.Collections.Generic;

namespace Mod
{
    public class Mod : MonoBehaviour
    {
        public const string ModTag = " [UT]";

        public static void Main()
        {
            ModAPI.RegisterPower<ReattachTool>(
                "Limb Reattachment Tool",
                "Reattaches the selected limb. Also repositions the limb even if it isn't dismembered.",
                ModAPI.LoadSprite("sprites/placeholder.png")
                );
            ModAPI.RegisterPower<ResetTool>( //Parented tools are in reverse order since that's how they show up on the powers menu.
                "Strength Reset Tool",
                "Resets the strength of the selected target. Does not work on gorses or androids. May cause weird behaviour if used on modded gorses.",
                ModAPI.LoadSprite("sprites/placeholder.png"),
                "Limb Reattachment Tool"
                );
            ModAPI.RegisterPower<ReviveTool>( //Parented tools are in reverse order since that's how they show up on the powers menu.
                "Revival Tool",
                "Barely revives targets in a radius.",
                ModAPI.LoadSprite("sprites/placeholder.png"),
                "Limb Reattachment Tool"
                );
            ModAPI.RegisterPower<DismemberTool>( //Parented tools are in reverse order since that's how they show up on the powers menu.
                "Dismembering Tool",
                "Dismembers the selected limb.",
                ModAPI.LoadSprite("sprites/placeholder.png"),
                "Limb Reattachment Tool"
                );
            /*ModAPI.RegisterPower<TestingTool>( //Parented tools are in reverse order since that's how they show up on the powers menu.
                "Testing Tool",
                "Unless you are me, you should not be seeing this. This is made for the explicit purpose of testing code.",
                ModAPI.LoadSprite("sprites/placeholder.png"),
                "Undismembering Tool"
                );*/

            Modification Regeneration = new Modification
            {
                OriginalItem = ModAPI.FindSpawnable("Pink Syringe"),
                NameOverride = "Natural Regeneration Syringe" + ModTag,
                DescriptionOverride = "Gives a genuinely natural regeneration. This is the closest you can get to perfection. Fairly slow, though.",
                CategoryOverride = ModAPI.FindCategory("Biohazard"),
                AfterSpawn = (Instance) =>
                {
                    Destroy(Instance.GetComponent<PinkSyringe>());
                    Instance.AddComponent<RegenerationSyringe>();
                }
            };
            ModAPI.RegisterLiquid(RegenerationSyringe.RegeneratingSerum.ID, new RegenerationSyringe.RegeneratingSerum());
            ModAPI.Register(Regeneration);

            Modification Revival = new Modification
            {
                OriginalItem = ModAPI.FindSpawnable("Pink Syringe"),
                NameOverride = "Revival Syringe" + ModTag,
                DescriptionOverride = "Gives the ability to revive from the dead, although only barely.",
                CategoryOverride = ModAPI.FindCategory("Biohazard"),
                AfterSpawn = (Instance) =>
                {
                    Destroy(Instance.GetComponent<PinkSyringe>());
                    Instance.AddComponent<RevivalSyringe>();
                }
            };
            ModAPI.RegisterLiquid(RevivalSyringe.RevivalSerum.ID, new RevivalSyringe.RevivalSerum());
            ModAPI.Register(Revival);
            ModAPI.OnItemSpawned += AddLimbLogger;
        }

        private static void AddLimbLogger(object sender, UserSpawnEventArgs args)
        {
            foreach (PersonBehaviour i in args.Instance.GetComponents<PersonBehaviour>())
            {
                foreach (LimbBehaviour i2 in i.Limbs)
                {
                    i2.gameObject.AddComponent<LimbLoggerBehaviour>();
                }
            }
        }

        [SkipSerialisation]
        public class LimbLoggerBehaviour : MonoBehaviour
        {
            private LimbBehaviour limb;
            public Vector2 Anchor { get; private set; }
            public Vector2 ConnectedAnchor { get; private set; }
            public Rigidbody2D ConnectedBody { get; private set; }
            public float BreakingThreshold { get; private set; }
            public bool UseLimits { get; private set; }
            public Vector3 Offset { get; private set; }
            public float BaseStrength { get; private set; }

            private void Start()
            {
                if (TryGetComponent(out limb) && limb.Joint)
                {
                    Anchor = limb.Joint.anchor;
                    ConnectedAnchor = limb.Joint.connectedAnchor;
                    ConnectedBody = limb.Joint.connectedBody;
                    BreakingThreshold = limb.BreakingThreshold;
                    UseLimits = limb.Joint.useLimits;
                    if (limb.transform.rotation.eulerAngles.z == 0f)
                    {
                        Offset = Utils.DivideVector3(limb.transform.position - limb.Joint.connectedBody.transform.position, limb.transform.root.localScale);
                    }
                    else
                    {
                        Offset = Vector3.zero;
                    }
                    BaseStrength = limb.BaseStrength;
                }
                else
                {
                    Destroy(this);
                }
            }
        }
    }

    public static class Utils
    {
        public static void RealignLimb(LimbBehaviour limb)
        {
            PersonBehaviour person = limb.GetComponentInParent<PersonBehaviour>();

            if (GetDescendantLimbs(limb) != null && person && limb)
            {
                foreach (LimbBehaviour i in GetDescendantLimbs(limb))
                {
                    if (i.TryGetComponent(out Mod.LimbLoggerBehaviour iLogger) && iLogger.ConnectedBody.TryGetComponent(out PhysicalBehaviour iConnectedPhysical))
                    {
                        if (!iConnectedPhysical.isDisintegrated)
                        {
                            i.transform.position = iLogger.ConnectedBody.transform.TransformPoint(iLogger.Offset);
                            i.transform.rotation = iLogger.ConnectedBody.transform.rotation;
                        }
                    }
                }
                foreach (LimbBehaviour i in person.Limbs)
                {
                    if (i.TryGetComponent(out Rigidbody2D rigidBody))
                    {
                        rigidBody.rotation = i.transform.eulerAngles.z;
                    }
                }
            }
        }

        public static Vector3 DivideVector3(Vector3 a, Vector3 b)
        {
            return new Vector3(!float.IsNaN(a.x / b.x) ? a.x / b.x : 0f, !float.IsNaN(a.y / b.y) ? a.y / b.y : 0f, !float.IsNaN(a.z / b.z) ? a.z / b.z : 0f);
        }

        public static bool IsConnectedToHead(LimbBehaviour limb)
        {
            LimbBehaviour currentLimb = limb;
            PersonBehaviour person = limb.GetComponentInParent<PersonBehaviour>();

            if (person && limb)
            {
                if (currentLimb.name == "Head")
                {
                    return true;
                }
                while (currentLimb.name != "UpperBody" && WhatToConnectTo(currentLimb, true) && !currentLimb.IsDismembered && currentLimb)
                {
                    if (currentLimb.name == "MiddleBody" && WhatToConnectTo(currentLimb, true).Joint.connectedBody.name != "MiddleBody")
                    {
                        return false;
                    }
                    currentLimb = WhatToConnectTo(currentLimb, true);
                }
                if (currentLimb.name == "UpperBody")
                {
                    foreach (LimbBehaviour i in person.Limbs)
                    {
                        if (i.name == "Head" && i.Joint.connectedBody.name == currentLimb.name)
                        {
                            return true;
                        }
                    }
                }
            }
            return false;
        }

        public static LimbBehaviour[] GetDescendantLimbs(LimbBehaviour limb)
        {
            PersonBehaviour person = limb.GetComponentInParent<PersonBehaviour>();
            List<LimbBehaviour> connectedLimbs = new List<LimbBehaviour>();
            List<LimbBehaviour> limbsToAdd = new List<LimbBehaviour>();
            int previousCount = 0;

            if (person && limb)
            {
                connectedLimbs.Add(limb);
                while (connectedLimbs.Count > previousCount)
                {
                    previousCount = connectedLimbs.Count;
                    foreach (LimbBehaviour i in person.Limbs)
                    {
                        if (i.TryGetComponent(out Mod.LimbLoggerBehaviour iLogger) && !i.IsDismembered)
                        {
                            foreach (LimbBehaviour i2 in connectedLimbs)
                            {
                                if (i2.name == iLogger.ConnectedBody.name)
                                {
                                    if (!connectedLimbs.Contains(i))
                                    {
                                        limbsToAdd.Add(i);
                                    }
                                }
                            }
                        }
                    }
                    foreach (LimbBehaviour i in limbsToAdd)
                    {
                        connectedLimbs.Add(i);
                    }
                    limbsToAdd.Clear();
                }
                return connectedLimbs.ToArray();
            }
            return null;
        }

        public static LimbBehaviour WhatToConnectTo(LimbBehaviour limb, bool bypassBackupCheck = false)
        {
            PersonBehaviour person = limb.GetComponentInParent<PersonBehaviour>();
            LimbBehaviour upperBody = null;
            if (limb.name == "UpperBody" || limb.GetComponent<PhysicalBehaviour>().isDisintegrated)
            {
                return null;
            }
            foreach (LimbBehaviour i in person.Limbs)
            {
                if (i.name == "UpperBody" && !i.GetComponent<PhysicalBehaviour>().isDisintegrated)
                {
                    upperBody = i;
                }
            }
            if (upperBody)
            {
                foreach (LimbBehaviour i in person.Limbs)
                {
                    if (limb == i || i.GetComponent<PhysicalBehaviour>().isDisintegrated)
                    {
                        continue;
                    }
                    if ((limb.name == "Head" || limb.name == "MiddleBody" || limb.name == "UpperArmFront" || limb.name == "UpperArm") && i == upperBody)
                    {
                        return i;
                    }
                    if (limb.name == "LowerBody" && i.name == "MiddleBody")
                    {
                        return i;
                    }
                    if ((limb.name == "UpperLegFront" || limb.name == "UpperLeg") && i.name == "LowerBody")
                    {
                        return i;
                    }
                    if (limb.name == "LowerLegFront" && i.name == "UpperLegFront")
                    {
                        return i;
                    }
                    if (limb.name == "LowerLeg" && i.name == "UpperLeg")
                    {
                        return i;
                    }
                    if (limb.name == "FootFront" && i.name == "LowerLegFront")
                    {
                        return i;
                    }
                    if (limb.name == "Foot" && i.name == "LowerLeg")
                    {
                        return i;
                    }
                    if (limb.name == "LowerArmFront" && i.name == "UpperArmFront")
                    {
                        return i;
                    }
                    if (limb.name == "LowerArm" && i.name == "UpperArm")
                    {
                        return i;
                    }
                }
                if (!bypassBackupCheck)
                {
                    return upperBody;
                }
            }
            return null;
        }

        public static bool IsConnectionGone(LimbBehaviour limb)
        {
            PersonBehaviour person = limb.GetComponentInParent<PersonBehaviour>();
            if (limb.GetComponent<PhysicalBehaviour>().isDisintegrated)
            {
                return false;
            }
            foreach (LimbBehaviour i in person.Limbs)
            {
                if (limb == i)
                {
                    continue;
                }
                if ((limb.name == "Head" || limb.name == "MiddleBody" || limb.name == "UpperArmFront" || limb.name == "UpperArm") && i.name == "UpperBody")
                {
                    return false;
                }
                if (limb.name == "LowerBody" && i.name == "MiddleBody")
                {
                    return false;
                }
                if ((limb.name == "UpperLegFront" || limb.name == "UpperLeg") && i.name == "LowerBody")
                {
                    return false;
                }
                if (limb.name == "LowerLegFront" && i.name == "UpperLegFront")
                {
                    return false;
                }
                if (limb.name == "LowerLeg" && i.name == "UpperLeg")
                {
                    return false;
                }
                if (limb.name == "FootFront" && i.name == "LowerLegFront")
                {
                    return false;
                }
                if (limb.name == "Foot" && i.name == "LowerLeg")
                {
                    return false;
                }
                if (limb.name == "LowerArmFront" && i.name == "UpperArmFront")
                {
                    return false;
                }
                if (limb.name == "LowerArm" && i.name == "UpperArm")
                {
                    return false;
                }
            }
            return true;
        }

        public static bool IsBeingStabbed(LimbBehaviour limb)
        {
            if (limb.TryGetComponent(out PhysicalBehaviour physical))
            {
                foreach (PhysicalBehaviour.Penetration i in physical.victimPenetrations)
                {
                    if (i.Active && i.Stabber && i.Stabber.StabCausesWound)
                    {
                        return true;
                    }
                }
                return false;
            }
            return false;
        }
    }

    public class ReattachTool : ToolBehaviour
    {
        public override void OnToolChosen() { }
        public override void OnToolUnchosen() { }
        public override void OnHold() { }
        public override void OnFixedHold() { }
        public override void OnDeselect() { }
        public override void OnSelect()
        {
            PhysicalBehaviour physical = ActiveSingleSelected;
            PersonBehaviour person = physical.GetComponentInParent<PersonBehaviour>();

            if (!physical.isDisintegrated && person && physical.TryGetComponent(out LimbBehaviour limb) && physical.TryGetComponent(out ConnectedNodeBehaviour node) && physical.TryGetComponent(out CirculationBehaviour circulation))
            {
                Utils.RealignLimb(limb);
                if (limb.IsDismembered || Utils.IsConnectionGone(limb))
                {
                    node.IsRoot = true;
                    limb.IsDismembered = false;
                    circulation.IsDisconnected = true;
                    if (limb.name != "MiddleBody" && !limb.HasJoint && limb.TryGetComponent(out Mod.LimbLoggerBehaviour logger) && logger.ConnectedBody.TryGetComponent(out CirculationBehaviour connectedCirculation) && logger.ConnectedBody.TryGetComponent(out ConnectedNodeBehaviour connectedNode) && logger.ConnectedBody.TryGetComponent(out LimbBehaviour connectedLimb))
                    {
                        limb.Joint = limb.gameObject.AddComponent<HingeJoint2D>();
                        limb.HasJoint = true;
                        limb.Joint.useMotor = true;
                        JointAngleLimits2D limits = limb.Joint.limits;
                        limits.min = limb.OriginalJointLimits.x;
                        limits.max = limb.OriginalJointLimits.y;
                        if (limb.Broken)
                        {
                            limits.max = Mathf.Lerp(limits.max, 180f, 0.5f);
                            limits.min = Mathf.Lerp(limits.min, -180f, 0.5f);
                        }
                        limb.Joint.limits = limits;
                        limb.Joint.anchor = logger.Anchor;
                        limb.Joint.connectedAnchor = logger.ConnectedAnchor;
                        limb.Joint.connectedBody = logger.ConnectedBody;
                        limb.gameObject.GetOrAddComponent<JointBreakFixerBehaviour>();
                        if (limb.CirculationBehaviour.BleedingRate == 10f)
                        {
                            limb.CirculationBehaviour.BleedingRate = 0f;
                        }
                        if (connectedCirculation.BleedingRate == 10f)
                        {
                            connectedCirculation.BleedingRate = 0f;
                        }
                        if (limb.HasBrain && connectedNode)
                        {
                            connectedNode.IsRoot = true;
                        }
                        if (circulation.WasInitiallyPumping && connectedNode && connectedCirculation)
                        {
                            connectedNode.IsRoot = true;
                            connectedLimb.IsDismembered = false;
                            connectedCirculation.IsDisconnected = true;
                            ForceCirculationBehaviour artificial = Utils.WhatToConnectTo(connectedLimb).gameObject.GetOrAddComponent<ForceCirculationBehaviour>();
                            if (!artificial.circulations.Contains(connectedCirculation) && artificial)
                            {
                                artificial.circulations.Add(connectedCirculation);
                            }
                        }
                        if (limb.BreakingThreshold < logger.BreakingThreshold)
                        {
                            limb.BreakingThreshold = logger.BreakingThreshold;
                        }
                        limb.Joint.useLimits = logger.UseLimits;
                    }
                }
                if (limb.name == "MiddleBody")
                {
                    if (!Utils.WhatToConnectTo(limb).NodeBehaviour.IsConnectedTo(node))
                    {
                        node.IsRoot = true;
                        limb.IsDismembered = false;
                        circulation.IsDisconnected = true;
                    }
                }
                if (circulation.IsDisconnected)
                {
                    ForceCirculationBehaviour artificial = Utils.WhatToConnectTo(limb).gameObject.GetOrAddComponent<ForceCirculationBehaviour>();
                    if (!artificial.circulations.Contains(circulation) && artificial)
                    {
                        artificial.circulations.Add(circulation);
                    }
                }
            }
        }

        private class JointBreakFixerBehaviour : MonoBehaviour
        {
            private void OnJointBreak2D(Joint2D joint)
            {
                if (TryGetComponent(out LimbBehaviour limb) && TryGetComponent(out CirculationBehaviour circulation) && TryGetComponent(out ConnectedNodeBehaviour node) && TryGetComponent(out Mod.LimbLoggerBehaviour logger) && logger.ConnectedBody.TryGetComponent(out ConnectedNodeBehaviour connectedNode) && logger.ConnectedBody.TryGetComponent(out CirculationBehaviour connectedCirculation) && !node.IsConnectedTo(connectedNode))
                {
                    if (joint != limb.Joint)
                    {
                        return;
                    }
                    if (circulation.IsDisconnected)
                    {
                        connectedCirculation.BleedingRate += 10f;
                    }
                    if (Utils.WhatToConnectTo(limb).TryGetComponent(out ForceCirculationBehaviour artificial))
                    {
                        if (artificial.circulations.Contains(circulation))
                        {
                            artificial.circulations.Remove(circulation);
                        }
                        if (artificial.circulations.Count <= 0)
                        {
                            Destroy(artificial);
                        }
                    }
                    else if (limb.name == "UpperBody" && limb.TryGetComponent(out artificial) && connectedCirculation)
                    {
                        if (artificial.circulations.Contains(connectedCirculation))
                        {
                            artificial.circulations.Remove(connectedCirculation);
                        }
                        if (artificial.circulations.Count <= 0)
                        {
                            Destroy(artificial);
                        }
                    }
                    if (limb.name != "Head")
                    {
                        node.IsRoot = false;
                        if (limb.name == "UpperBody")
                        {
                            connectedNode.IsRoot = false;
                        }
                    }
                    else
                    {
                        connectedNode.IsRoot = false;
                    }
                }
            }
        }
    }

    public class ForceCirculationBehaviour : MonoBehaviour
    {
        public List<CirculationBehaviour> circulations = new List<CirculationBehaviour>();
        PersonBehaviour person;
        LimbBehaviour limb;
        CirculationBehaviour circulation;

        private void Awake()
        {
            circulation = GetComponent<CirculationBehaviour>();
            person = GetComponentInParent<PersonBehaviour>();
            limb = GetComponent<LimbBehaviour>();
        }

        private void FixedUpdate()
        {
            if (circulation && person && limb)
            {
                float amount = Time.fixedDeltaTime;
                foreach (CirculationBehaviour i in circulations)
                {
                    if (!i.Limb.PhysicalBehaviour.isDisintegrated && i.IsDisconnected)
                    {
                        if (i.TotalLiquidAmount < circulation.TotalLiquidAmount && circulation.TotalLiquidAmount > circulation.LowerLimit && i.TotalLiquidAmount < i.UpperLimit)
                        {
                            circulation.TransferTo(amount, i, true);
                            i.DeleteEmptyLiquidEntries();
                        }
                        if (circulation.TotalLiquidAmount < i.TotalLiquidAmount && i.TotalLiquidAmount > i.LowerLimit && circulation.TotalLiquidAmount < circulation.UpperLimit)
                        {
                            i.TransferTo(amount, circulation, true);
                            circulation.DeleteEmptyLiquidEntries();
                        }
                    }
                    if (!i.Limb.PhysicalBehaviour.isDisintegrated && i.BloodFlow < circulation.BloodFlow && i.IsDisconnected)
                    {
                        i.BloodFlow = Mathf.Lerp(i.BloodFlow, circulation.BloodFlow, 0.5f);
                    }
                }
            }
        }
    }

    public class DismemberTool : ToolBehaviour
    {
        public override void OnToolChosen() { }
        public override void OnToolUnchosen() { }
        public override void OnHold() { }
        public override void OnFixedHold() { }
        public override void OnDeselect() { }
        public override void OnSelect()
        {
            PhysicalBehaviour physical = ActiveSingleSelected;

            if (physical.TryGetComponent(out LimbBehaviour limb) && physical.TryGetComponent(out ConnectedNodeBehaviour node))
            {
                if (!limb.HasBrain)
                {
                    node.IsRoot = false;
                }
                if (limb.Joint)
                {
                    limb.BreakingThreshold = 0f;
                }
                else
                {
                    limb.IsDismembered = true;
                }
            }
        }
    }

    public class ReviveTool : AOEPowerTool
    {
        protected override void HandleObject(PhysicalBehaviour physical)
        {
            PersonBehaviour person = physical.GetComponentInParent<PersonBehaviour>();

            if (person && physical.TryGetComponent(out LimbBehaviour limb) && physical.TryGetComponent(out CirculationBehaviour circulation) && physical.TryGetComponent(out SkinMaterialHandler skinHandler))
            {
                if ((circulation.WasInitiallyPumping || limb.HasBrain) && Utils.IsConnectedToHead(limb))
                {
                    circulation.IsPump = circulation.WasInitiallyPumping;
                    person.Braindead = false;
                    person.BrainDamaged = false;
                    person.BrainDamagedTime = 0f;
                    limb.LungsPunctured = false;
                    if (limb.Health < limb.InitialHealth * 0.05f)
                    {
                        limb.Health = limb.InitialHealth * 0.05f;
                    }
                    if (physical.BurnProgress > 0.9f)
                    {
                        physical.BurnProgress = 0.9f;
                        circulation.AddLiquid(limb.GetOriginalBloodType(), 1f - circulation.GetAmountOfBlood());
                    }
                    if (skinHandler.AcidProgress > 0.9f)
                    {
                        skinHandler.AcidProgress = 0.9f;
                        circulation.AddLiquid(limb.GetOriginalBloodType(), 1f - circulation.GetAmountOfBlood());
                    }
                    if (circulation.GetAmountOfBlood() < 0.4f)
                    {
                        circulation.AddLiquid(limb.GetOriginalBloodType(), 0.4f - circulation.GetAmountOfBlood());
                    }
                    if (person.Consciousness < 0.1f)
                    {
                        person.Consciousness = 0.1f;
                    }
                    if (person.OxygenLevel < 0.1f)
                    {
                        person.OxygenLevel = 0.1f;
                    }
                }
            }

        }

        public override float Radius
        {
            get
            {
                return 0.5f;
            }
        }

        public override float DampeningRadius
        {
            get
            {
                return 0.5f;
            }
        }
    }

    public class ResetTool : ToolBehaviour
    {
        public override void OnToolChosen() { }
        public override void OnToolUnchosen() { }
        public override void OnHold() { }
        public override void OnFixedHold() { }
        public override void OnDeselect() { }
        public override void OnSelect()
        {
            PhysicalBehaviour physical = ActiveSingleSelected;
            PersonBehaviour person = physical.GetComponentInParent<PersonBehaviour>();

            if (person && physical.TryGetComponent(out LimbBehaviour limb) && physical.TryGetComponent(out Mod.LimbLoggerBehaviour logger))
            {
                if (person.name != "Gorse" && logger.BaseStrength > 0f)
                {
                    limb.BaseStrength = logger.BaseStrength;
                }
            }
        }
    }

    public class TestingTool : ToolBehaviour
    {
        public override void OnToolChosen() { }
        public override void OnToolUnchosen() { }
        public override void OnHold() { }
        public override void OnFixedHold() { }
        public override void OnDeselect() { }
        public override void OnSelect() { }
    }

    public class RegenerationSyringe : SyringeBehaviour
    {
        public override string GetLiquidID() => RegeneratingSerum.ID;

        public class RegeneratingSerum : Liquid
        {
            public const string ID = "REGENERATION SERUM" + Mod.ModTag;

            public RegeneratingSerum()
            {
                Color = new Color(1f, 0.5f, 0.5f);
            }

            public override void OnEnterLimb(LimbBehaviour limb)
            {
                limb.gameObject.GetOrAddComponent<RegenBehaviour>();
            }

            public override void OnUpdate(BloodContainer container)
            {
                if (container is CirculationBehaviour)
                {
                    container.RemoveLiquid(this, Mathf.Max(0.1f, container.GetAmount(this) * 0.1f));
                }
            }
            public override void OnEnterContainer(BloodContainer container) { }
            public override void OnExitContainer(BloodContainer container) { }
        }
    }

    [SkipSerialisation]
    public class RegenBehaviour : MonoBehaviour
    {
        PersonBehaviour person;
        LimbBehaviour limb;
        Regen regen;

        public void Start()
        {
            person = GetComponentInParent<PersonBehaviour>();
            limb = GetComponent<LimbBehaviour>();
        }

        private void Update()
        {
            if (limb && person)
            {
                if (person.IsAlive() && Utils.IsConnectedToHead(limb) && !regen)
                {
                    regen = gameObject.GetOrAddComponent<Regen>();
                }
            }
        }
    }

    [SkipSerialisation]
    public class Regen : MonoBehaviour
    {
        CirculationBehaviour circulation;
        LimbBehaviour limb;
        PersonBehaviour person;
        SkinMaterialHandler skinHandler;
        PhysicalBehaviour physical;
        public float healTimer = 0f;
        private float boneTimer = 0f;
        private float bruiseTimer = 0f;
        private float gunTimer = 0f;
        private float stabTimer = 0f;
        private float lungTimer = 0f;
        private float heartTimer = 0f;
        private float bleedTimer = 0f;
        private float internalBleedTimer = 0f;
        private float bloodTimer = 0f;
        private float previousBleedingRate = 0f;
        private float previousHealth = 0f;
        private float healthMultiplier = 0f;
        private float brainTimer = 0f;
        private int damagePointCount = 0;
        private bool healTimerStarted = false;

        public void Start()
        {
            circulation = GetComponent<CirculationBehaviour>();
            limb = GetComponent<LimbBehaviour>();
            person = GetComponentInParent<PersonBehaviour>();
            skinHandler = GetComponent<SkinMaterialHandler>();
            physical = GetComponent<PhysicalBehaviour>();
            limb.RegenerationSpeed = 0f;
            if (limb)
            {
                previousHealth = limb.Health;
            }
        }

        private void Update()
        {
            healthMultiplier = Mathf.Clamp(limb.Health, 0f, limb.InitialHealth) / limb.InitialHealth;
            if (!circulation || !limb || !person || !skinHandler || !physical || !person.IsAlive() || !Utils.IsConnectedToHead(limb))
            {
                Destroy(this);
            }
            if (person.Consciousness < 1f)
            {
                person.Consciousness = Mathf.Min(1f, person.Consciousness + Time.deltaTime * 0.01f * circulation.BloodFlow * (person.AdrenalineLevel * 2f + 1f));
            }
            if (person.ShockLevel > 0f)
            {
                person.ShockLevel = Mathf.Max(0f, person.ShockLevel - Time.deltaTime * 0.01f * (person.AdrenalineLevel * 2f + 1f));
            }
            if (limb.Numbness > 0f)
            {
                limb.Numbness = Mathf.Max(0f, limb.Numbness - Time.deltaTime * 0.01f * circulation.BloodFlow / (person.AdrenalineLevel * 2f + 1f));
            }
            if (circulation.BleedingRate > 0f)
            {
                if (previousBleedingRate != 0f && previousBleedingRate < circulation.BleedingRate)
                {
                    bleedTimer *= 0.5f;
                }
                bleedTimer = Mathf.Min(5f, bleedTimer + Time.deltaTime * 0.2f);
                circulation.BleedingRate = Mathf.Clamp(Mathf.Max(0f, circulation.BleedingRate - Mathf.Max(Time.deltaTime * 0.2f * Mathf.Max(0.1f, healthMultiplier) / Mathf.Clamp((circulation.StabWoundCount + circulation.GunshotWoundCount) * 0.1f, 1f, 10f) * (1f + circulation.BleedingPointCount * Mathf.Min(1f, circulation.BleedingPointCount * 0.01f)), Time.deltaTime * Mathf.Lerp(circulation.BleedingRate, 0f, 0.04f) * Mathf.Max(0.1f, healthMultiplier) / Mathf.Clamp((circulation.StabWoundCount + circulation.GunshotWoundCount) * 0.1f, 1f, 10f) * (1f + circulation.BleedingPointCount * Mathf.Min(1f, circulation.BleedingPointCount * 0.01f)) * bleedTimer)), 0f, 300f);
                previousBleedingRate = circulation.BleedingRate;
            }
            else
            {
                foreach (GameObject i in circulation.BleedingParticles)
                {
                    Destroy(i);
                }
                circulation.BleedingParticles.Clear();
                circulation.BleedingPointCount = 0;
                bleedTimer = Mathf.Max(0f, bleedTimer - Time.deltaTime * 0.5f);
                previousBleedingRate = 0f;
            }
            if (circulation.InternalBleedingIntensity > 0f)
            {
                internalBleedTimer = Mathf.Min(5f, internalBleedTimer + Time.deltaTime * 0.2f);
                circulation.InternalBleedingIntensity = Mathf.Max(0f, circulation.InternalBleedingIntensity - Mathf.Max(Time.deltaTime * 0.2f * Mathf.Max(0.2f, healthMultiplier) / Mathf.Clamp(limb.BruiseCount * 0.1f, 1f, 10f), Time.deltaTime * Mathf.Lerp(circulation.InternalBleedingIntensity, 0f, 0.04f) * Mathf.Max(0.2f, healthMultiplier) / Mathf.Clamp(limb.BruiseCount * 0.1f, 1f, 10f) * internalBleedTimer));
            }
            else
            {
                internalBleedTimer = Mathf.Max(0f, internalBleedTimer - Time.deltaTime * 0.5f);
            }
            if (physical.BurnProgress > 0f)
            {
                physical.BurnProgress = Mathf.Clamp(physical.BurnProgress - Time.deltaTime * 0.02f, 0f, 1f);
            }
            if (skinHandler.AcidProgress > 0f)
            {
                skinHandler.AcidProgress = Mathf.Clamp(skinHandler.AcidProgress - Time.deltaTime * 0.02f, 0f, 1f);
            }
            if (skinHandler.RottenProgress > 0f && !limb.IsZombie)
            {
                skinHandler.RottenProgress = Mathf.Clamp(skinHandler.RottenProgress - Time.deltaTime * 0.02f * Mathf.Clamp(limb.Health, 0f, limb.InitialHealth), 0f, 1f);
            }
            if (circulation.GetAmountOfBlood() < 1f && healTimer <= 0f)
            {
                bloodTimer = Mathf.Min(10f, bloodTimer + Time.deltaTime * 0.2f / Mathf.Min(20f, (circulation.StabWoundCount + circulation.GunshotWoundCount + limb.BruiseCount) * circulation.BloodFlow));
                circulation.AddLiquid(limb.GetOriginalBloodType(), Mathf.Min(1f - circulation.GetAmountOfBlood(), Time.deltaTime * 0.003f * healthMultiplier) * bloodTimer);
            }
            else
            {
                bloodTimer = Mathf.Max(0f, bloodTimer - Time.deltaTime * 0.25f * Mathf.Max(1f, bloodTimer));
            }
            if (limb.Health >= limb.InitialHealth)
            {
                healTimerStarted = false;
            }
            else if (!healTimerStarted && healTimer <= 0f)
            {
                healTimer = Random.Range(1f, 3f);
                SyncHealTimers(0.6f, true);
                healTimerStarted = true;
            }
            if (healTimer > 0f)
            {
                healTimer -= Time.deltaTime * Mathf.Max(0.2f, healthMultiplier);
            }
            else
            {
                if (previousHealth > 0f && previousHealth > limb.Health + limb.InitialHealth * 0.3f)
                {
                    healTimer = Random.Range(0.5f, 2f);
                    SyncHealTimers(0.3f, true);
                }
                else if (previousHealth > 0f && previousHealth > limb.Health + limb.InitialHealth * 0.1f && limb.Health <= 0f)
                {
                    healTimer = Random.Range(0.5f, 1.5f);
                    SyncHealTimers(0.2f, true);
                }
                limb.Health = Mathf.Clamp(limb.Health + limb.InitialHealth * Time.deltaTime * 0.05f * circulation.BloodFlow, 0f, limb.InitialHealth);
                previousHealth = limb.Health;
            }
            for (int i = 0; i < skinHandler.damagePoints.Length; i++)
            {
                Vector4[] damagePointArray = skinHandler.damagePoints;
                DamagePoint damagePoint = damagePointArray[i];
                damagePointArray[i] = new DamagePoint(damagePoint.X, damagePoint.Y, Mathf.Clamp(damagePoint.Intensity - Time.deltaTime * 0.5f * healthMultiplier, 0f, 100f), damagePoint.Type);
                skinHandler.damagePoints = damagePointArray;
                if (damagePoint.Intensity > 0f)
                {
                    damagePointCount += 1;
                }
            }
            skinHandler.currentDamagePointCount = damagePointCount;
            damagePointCount = 0;
            skinHandler.Sync();
            if (bruiseTimer < 1f && limb.BruiseCount > 0f)
            {
                bruiseTimer += Time.deltaTime * 3f * healthMultiplier * (1f + limb.BruiseCount * 0.01f);
            }
            else if (limb.BruiseCount > 0f)
            {
                limb.BruiseCount = (ushort)Mathf.Max(0, limb.BruiseCount - (int)bruiseTimer);
                bruiseTimer = Mathf.Min(limb.BruiseCount, bruiseTimer - (int)bruiseTimer);
            }
            if (gunTimer < 1f && circulation.GunshotWoundCount > 0)
            {
                gunTimer += Time.deltaTime * 3f * healthMultiplier * (1f + circulation.GunshotWoundCount * 0.01f);
            }
            else if (circulation.GunshotWoundCount > 0)
            {
                circulation.GunshotWoundCount = (ushort)Mathf.Max(0, circulation.GunshotWoundCount - (int)gunTimer);
                gunTimer = Mathf.Min(circulation.GunshotWoundCount, gunTimer - (int)gunTimer);
            }
            if (stabTimer < 1f && circulation.StabWoundCount > 0)
            {
                stabTimer += Time.deltaTime * 3f * healthMultiplier * (1f + circulation.StabWoundCount * 0.01f);
            }
            else if (circulation.StabWoundCount > 0)
            {
                circulation.StabWoundCount = (ushort)Mathf.Max(0, circulation.StabWoundCount - (int)stabTimer);
                stabTimer = Mathf.Min(circulation.StabWoundCount, stabTimer - (int)stabTimer);
            }
            if (limb.Broken && boneTimer <= 0f)
            {
                boneTimer = Random.Range(2.5f, 7.5f);
            }
            else if (limb.Broken)
            {
                boneTimer -= Time.deltaTime * healthMultiplier;
                if (boneTimer <= 0f)
                {
                    limb.HealBone();
                }
            }
            if (limb.LungsPunctured && lungTimer <= 0f)
            {
                lungTimer = Random.Range(5f, 15f);
            }
            else if (limb.LungsPunctured)
            {
                lungTimer -= Time.deltaTime * Mathf.Clamp(person.OxygenLevel, 0.1f, 1f);
                if (lungTimer <= 0f)
                {
                    limb.LungsPunctured = false;
                }
            }
            if (circulation.WasInitiallyPumping)
            {
                if (!circulation.IsPump && heartTimer <= 0f)
                {
                    heartTimer = Random.Range(5f, 12f);
                }
                else if (!circulation.IsPump)
                {
                    heartTimer -= Time.deltaTime * Mathf.Clamp(circulation.BloodFlow, 0.1f, 1f);
                    if (heartTimer <= 0f)
                    {
                        circulation.IsPump = true;
                    }
                }
                else if (heartTimer >= 0f)
                {
                    heartTimer = 0f;
                }
            }
            if (limb.HasBrain)
            {
                if (person.BrainDamaged && brainTimer <= 0f)
                {
                    brainTimer = Random.Range(20f, 40f);
                }
                else if (person.BrainDamaged)
                {
                    brainTimer -= Time.deltaTime;
                    if (brainTimer <= 0f)
                    {
                        person.BrainDamaged = false;
                        person.BrainDamagedTime = 0f;
                    }
                }
                else if (brainTimer >= 0f)
                {
                    brainTimer = 0f;
                }
            }
        }

        private void SyncHealTimers(float multiplier = 0.5f, bool random = false)
        {
            float finalMultiplier = multiplier;
            foreach (Regen i in person.GetComponentsInChildren<Regen>())
            {
                if (random)
                {
                    finalMultiplier = Random.Range(0f, multiplier);
                }
                if (i.healTimer < healTimer * finalMultiplier)
                {
                    i.healTimer = healTimer * finalMultiplier;
                }
            }
        }
    }

    public class RevivalSyringe : SyringeBehaviour
    {
        public override string GetLiquidID() => RevivalSerum.ID;

        public class RevivalSerum : Liquid
        {
            public const string ID = "REVIVAL SERUM" + Mod.ModTag;

            public RevivalSerum()
            {
                Color = new Color(0.3f, 0.8f, 0.8f);
            }

            public override void OnEnterLimb(LimbBehaviour limb)
            {
                if (limb.name == "Head" || limb.name == "UpperBody")
                {
                    limb.gameObject.GetOrAddComponent<ReviveBehaviour>();
                }
            }

            public override void OnUpdate(BloodContainer container)
            {
                if (container is CirculationBehaviour)
                {
                    container.RemoveLiquid(this, Mathf.Max(0.1f, container.GetAmount(this) * 0.1f));
                }
            }
            public override void OnEnterContainer(BloodContainer container) { }
            public override void OnExitContainer(BloodContainer container) { }
        }
    }

    [SkipSerialisation]
    public class ReviveBehaviour : MonoBehaviour
    {
        CirculationBehaviour circulation;
        LimbBehaviour limb;
        PersonBehaviour person;
        SkinMaterialHandler skinHandler;
        PhysicalBehaviour physical;
        private bool canHealBlood = false;
        private float deathTimer = 0f;

        private void Start()
        {
            circulation = GetComponent<CirculationBehaviour>();
            limb = GetComponent<LimbBehaviour>();
            person = GetComponentInParent<PersonBehaviour>();
            skinHandler = GetComponent<SkinMaterialHandler>();
            physical = GetComponent<PhysicalBehaviour>();
        }

        private void Update()
        {
            if (limb && circulation && person && skinHandler && physical)
            {
                if (person.Braindead)
                {
                    person.Braindead = false;
                }
                if (limb.name == "UpperBody" || limb.name == "Head")
                {
                    if (limb.name == "Head" && limb.Joint.connectedBody.name != "UpperBody")
                    {
                        return;
                    }
                    if (!person.IsAlive() && !IsAnyReviveBeingStabbed() && Utils.IsConnectedToHead(limb) && !physical.isDisintegrated)
                    {
                        deathTimer += Time.deltaTime;
                        limb.Health = Mathf.Max(limb.Health, limb.InitialHealth * 0.05f);
                        limb.LungsPunctured = false;
                        circulation.IsPump = circulation.WasInitiallyPumping;
                        physical.BurnProgress = Mathf.Clamp(physical.BurnProgress, 0f, 0.9f);
                        skinHandler.AcidProgress = Mathf.Clamp(skinHandler.AcidProgress, 0f, 0.9f);
                        if (circulation.GetAmountOfBlood() <= 0f)
                        {
                            canHealBlood = true;
                        }
                        else if (circulation.GetAmountOfBlood() <= 0.4f && deathTimer >= 5f)
                        {
                            canHealBlood = true;
                        }
                    }
                    else
                    {
                        deathTimer = 0f;
                    }
                    if (circulation.GetAmountOfBlood() >= 0.5f)
                    {
                        canHealBlood = false;
                    }
                    if (circulation.GetAmountOfBlood() < 0.4f && canHealBlood && !IsAnyReviveBeingStabbed())
                    {
                        circulation.AddLiquid(limb.GetOriginalBloodType(), 0.4f - circulation.GetAmountOfBlood());
                    }
                }
            }
            else
            {
                Destroy(this);
            }
        }

        private bool IsAnyReviveBeingStabbed()
        {
            if (person)
            {
                foreach (LimbBehaviour i in person.Limbs)
                {
                    if (i.GetComponent<ReviveBehaviour>())
                    {
                        if (Utils.IsBeingStabbed(i))
                        {
                            return true;
                        }
                    }
                }
            }
            return false;
        }
    }
}